
from werkzeug.exceptions import *
from werkzeug.exceptions import abort

from .routing import *
from .context import request, g, G
from .application import Application as application, response

def run(view_func, **kwargs):
	"""runs a single route application"""
	application(route('/', view_func)).run(**kwargs)




from dataclasses import dataclass

from werkzeug.local import Local, LocalManager, LocalProxy, LocalStack

local = Local()
local_manager = LocalManager([local])
request = local('request')
g = local('g')

class G(object):
    """user managed request context
    """

    def __contains__(self, key):
        return hasattr(self, key)




from dataclasses import dataclass, field, replace
from typing import Callable, Iterable
from functools import partial, wraps

from more_itertools import collapse, partition
from werkzeug.routing import Rule

__all__ = (
	'routes',
	'route',
	'get', 'post', 'put', 'delete', 'patch', 'options',
	'index', 'new', 'create', 'show', 'edit', 'update', 'destroy'
)


@dataclass
class Route():
	path: str
	view_func: Callable
	methods: list = field(default_factory=list)
	defaults: dict = field(default_factory=dict)
	middleware: list = field(default_factory=list)
	strict_slashes: bool = field(default_factory=lambda: None)

	def as_rule(self, **kwargs) -> Rule:
		return Rule(
			self.path,
			methods = self.methods or None,
			defaults = self.defaults,
			**kwargs
		)

	def mount(self, path=None):
		return replace(self, path = f'{path}{self.path}' if path else self.path)

	def insert_middleware(self, middleware):
		return replace(self, middleware = [*middleware, *self.middleware])

	def add_methods(self, methods):
		return replace(self, methods = list(set([*methods, *self.methods])))

	def add_defaults(self, defaults):
		return replace(self, defaults = {**defaults, **self.defaults})


def route(*args, **kwargs) -> Route:
	"""
	"""
	if len(args) == 1:
		view_func = args[0]
		path = f'/{view_func.__name__}'
	elif len(args) == 2:
		path, view_func = args
	else:
		raise ValueError(f'route() expected 1-2 args, got {len(args)}')
	return Route(
		path = path,
		view_func = view_func,
		**kwargs
	)


# some standard http methods
get = partial(route, methods=['GET'])
post = partial(route, methods=['POST'])
put = partial(route, methods=['PUT'])
delete = partial(route, methods=['DELETE'])
patch = partial(route, methods=['PATCH'])
options = partial(route, methods=['OPTIONS'])

# the 7 RESTful routes
index = partial(get, '/')
new = partial(get, '/new')
create = partial(post, '/')
show = partial(get, '/<id>')
edit = partial(get, '<id>/edit')
update = partial(route, '/<id>', methods=['PATCH', 'PUT'])
destroy = partial(delete, '/<id>')



def _routes(
	*args,
	path: str = None,
	middleware: list = [],
	methods: list = [],
	defaults: dict = {}
):
	"""DRY: create routes with similar configuration"""
	for route in collapse(args):
		yield route\
			.insert_middleware(middleware)\
			.mount(path)\
			.add_methods(methods)\
			.add_defaults(defaults)

def routes(*args, **kwargs) -> Iterable[Route]:
	"""DRY: create routes with similar configuration"""
	urls = []
	args = tuple(collapse(args))
	key = lambda arg: not isinstance(arg, Route)
	for i, arg in enumerate(args):
		if isinstance(arg, Route):
			urls.append(
				replace(
					arg,
					middleware = [
						*filter(key, args[:i]),
						*arg.middleware,
						*filter(key, args[i+1:])
					]
				)
			)
	return tuple(_routes(urls, **kwargs))




from dataclasses import replace
from typing import Callable, List
from functools import partial
from collections import OrderedDict

from more_itertools import collapse, partition
from werkzeug.routing import Map
from werkzeug.serving import run_simple
from werkzeug.exceptions import HTTPException
from werkzeug.wrappers import Request, Response as response

from .context import local, g, local_manager, G, request
from .routing import routes, Route
from .middleware import create_stack, notify_stack, Stack


def abort_middleware():
	"""support for abort()"""
	global HTTPException
	try:
		yield
	except HTTPException as response:
		yield response


def create_app(
	*args,
	g_factory: Callable = G,
	strict_slashes = False,
	request_class = Request
) -> Callable:
	"""convert routes into a WSGI application
	middleware added here is initilized BEFORE routing, so it can catch NotFound
	"""
	global Map, local_manager

	middleware, args = partition(
		lambda obj: isinstance(obj, Route),
		collapse(args)
	)

	endpoints = OrderedDict(enumerate(routes(*args)))
	rules = [
		route.as_rule(
			endpoint = i,
			strict_slashes = strict_slashes if route.strict_slashes is None else route.strict_slashes
		)
		for i, route in endpoints.items()
	]
	url_map = Map(rules)

	middleware = [
		abort_middleware,
		#force_type_middleware,
		*middleware,
		#force_type_middleware
	]

	@local_manager.make_middleware
	def application(environ, start_response):
		global local, force_type_middleware, Stack, Request
		nonlocal middleware, url_map, endpoints, g_factory, request_class

		print('NEW REQUEST', environ)

		# initialize thread locals
		environ['madness.request'] = local.request = request_class(environ)
		environ['madness.stack'] = stack = Stack()
		environ['madness.g'] = local.g = g_factory()

		try:

			try:
				# route the request, may raise HTTPException such as NotFound
				urls = url_map.bind_to_environ(environ)
				endpoint, kwargs = urls.match()

			except HTTPException as exception:
				# initialize the stack with application middleware
				stack.add(middleware)
				raise exception

			else:
				# apply defaults
				for key, value in kwargs.items():
					setattr(local.g, key, value)

				# initialize the stack with application middleware
				stack.add(middleware)

				route = endpoints[endpoint]

				# add the route's middleware to the stack
				stack.add(route.middleware)

				response = route.view_func()

		# allow the stack to swap out the response application
		except Exception as exception:
			response = stack.throw(exception)
		else:
			response = stack.send(response)

		# allow stack to handle errors that happen while streaming the response
		try:
			yield from response(environ, start_response)
		except Exception as exception:
			stack.throw(exception)
		else:
			stack.send(None)

	return application




class Application():
	""""""

	def __init__(self, *args, **kwargs):
		super().__init__()
		self.wsgi_app = create_app(*args, **kwargs)

	def __call__(self, environ, start_response):
		return self.wsgi_app(environ, start_response)

	def run(
		self,
		host: str = '127.0.0.1',
		port: int = 9090,
		debug: bool = True
	):
		run_simple(
			host,
			port,
			self.wsgi_app,
			use_reloader = debug,
			use_debugger = False,
			use_evalex = False,
		)




from inspect import isgenerator
from typing import Iterable, Callable, Generator, Any


class Stack(list):

	def add(self, middleware: Iterable[Callable]):
		for gen in create_stack(middleware):
			self.append(gen)

	def send(self, value: Any):
		return notify_stack(self, value = value)

	def throw(self, exception: Exception):
		return notify_stack(self, exception = exception)


def create_stack(
	middleware: Iterable[Callable]
) -> Iterable[Generator]:
	for func in middleware:
		gen = func()
		if isgenerator(gen):
			gen.send(None)
			yield gen

def notify_stack(
	stack: Iterable[Generator],
	value: Any = None,
	exception: Exception = None
):
	"""
    Errors should never pass silently.
    Unless explicitly silenced.
    """
	for gen in reversed(stack):
		try:
			if exception != None:
				value2 = gen.throw(exception)
			else:
				value2 = gen.send(value)
		except StopIteration:
			exception = None
		except Exception as new_exception:
			exception = new_exception
			value = None
		else:
			exception = None
			if value2 != None:
				value = value2
	if exception != None:
		raise exception
	return value




from setuptools import setup, find_packages
from pathlib import Path

setup(
    name = "madness",
    version = "0.7.0",
    author = "Forrest Button",
    author_email = "forrest.button@gmail.com",
    description = "wsgi microframework suitable for building modular DRY RESTful APIs",
    long_description = Path("README.md").read_text(),
    long_description_content_type = "text/markdown",
    url = "https://github.com/Waffles32/madness",
    packages = ['madness'],
    classifiers = [
        "Development Status :: 2 - Pre-Alpha",
        "Topic :: Internet :: WWW/HTTP :: WSGI",
        "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
	include_package_data = False,
	install_requires = Path('requirements.txt').read_text().splitlines(),
	python_requires='>3.6',
	zip_safe = True,
)
